/* =========================================================================
	This library is part of IshiSoft application.
	------------------------------------------------------------------------
	author: Yuichiro Ishitani
			http://www001.upp.so-net.ne.jp/ishi/index.html
  ========================================================================== */
#include "memory.h"
#include "MyHashEx.h"

#if !defined(size_t)
#define size_t unsigned int
#endif

// | KEY |\0|VALUE|\0| KEY |\0|VALUE|\0\0 <- terminator

static char	szEmpty[] = "";


static size_t StrLen( const char *string )
{
	size_t ii = 0;

	while( *string++ != '\0' )
		ii++;
	
	return ii;
}

//-----------------------------------------------------------------------------
// Perform a lowercase comparison of strings.
static int StrCmpI( const char *str1, const char *str2 )
{
	int a, b;

	do
	{
		a = (unsigned char)*str1;
		b = (unsigned char)*str2;

		if( (a >= 'A') && (a <= 'Z') )
			a += 'a' - 'A';

		if( (b >= 'A') && (b <= 'Z') )
			b += 'a' - 'A';

		str1++;
		str2++;
	}while( a && (a == b) );
	
	return a - b;
}


//-----------------------------------------------------------------------------
// Compare strings
static int StrCmp( const char *str1, const char *str2 )
{
	while( *str1 && (*str1 == *str2) )
	{
		str1++;
		str2++;
	}
	
	return *str1 - *str2;
}

//-----------------------------------------------------------------------------
//	@detail
//		Constracter
//	@param
//	@return
//-----------------------------------------------------------------------------
MyHashEx::MyHashEx()
{
	m_pHeap = new MyVarMem;
}

MyHashEx::~MyHashEx()
{
	if( m_pHeap )
		delete m_pHeap;
}

//-----------------------------------------------------------------------------
//	@detail
//		Add record(key and value).
//	@param
//		cahr*	pszKey		Key
//		char*	pszValue	value
//	@return
//		Number of add bytes to heap.
//-----------------------------------------------------------------------------
int MyHashEx::Set(const char* pszKey, const char* pszValue)
{
	if( NULL == pszKey || NULL == pszValue )
		return 0;

	int ret;
	int keylen = StrLen(pszKey) + 1;	// obtain null terminate.
	int valuelen = StrLen(pszValue) + 1;
	char* record = (char*)MALLOC(keylen + valuelen);

	if( NULL == record )
		return 0;

	memset(record, 0x00,	keylen + valuelen);
	memcpy(record,			pszKey,		keylen);
	memcpy(record + keylen,	pszValue,	valuelen);


	// If a key already exists, delete it.
	Del(pszKey);

	// -- 

	
	ret = m_pHeap->Add(record, keylen + valuelen);

	FREEIF( record );
	return ret;
}

//-----------------------------------------------------------------------------
//	@detail
//		Add record(key and value).
//	@param
//		cahr*	pszKey		Key
//		char*	pszValue	value
//	@return
//		Number of add bytes to heap.
//-----------------------------------------------------------------------------
int MyHashEx::Set(const char* pszKey, unsigned long dwValue)
{
	if( NULL == pszKey )
		return 0;

	int ii;
	unsigned long val = dwValue;
	char szNum[16] = {0};
	for( ii = 0; ii < 11; ii++ )
	{
		szNum[15-ii] = (char)(val %= 10);
		if( 0 == val )
			break;
	}

	return Set( pszKey, &szNum[ii] );
	
}

//-----------------------------------------------------------------------------
//	@detail
//		Delete record
//	@param
//		char* pszKey	Key to delete
//	@return
//		success	1
//		fail	0
//	@comment
//		Deletion is caused rearrangement of memory.
//		If possible don't use this function.
//-----------------------------------------------------------------------------
int MyHashEx::Del(const char* pszKey)
{
	char* pTop = (char*)m_pHeap->GetBuffer();
	const char* pDelKey = SeekToKeyI(pszKey);

	// no key.
	if( NULL == pDelKey )
	{
		return 0;
	}

	int nTotalLen = m_pHeap->GetLength();

	// the key at the head.
	if(pTop == pDelKey)
	{
		MyVarMem* pHead = new MyVarMem;
		if( NULL == pHead )
		{
			return 0;
		}

		const char* pNew = SeekNext(pDelKey);
		
		if( NULL != pNew )
		{
			pHead->Add(pNew, nTotalLen - (pNew - pTop));
		}
		delete m_pHeap;
		m_pHeap = pHead;
		SeekToFirst();
		return 1;
	}

	// there are in the tail end.
	const char* pNext = SeekNext(pDelKey);

	if( NULL == pNext )
	{
		MyVarMem* pHead = new MyVarMem;
		if( NULL == pHead )
		{
			return 0;
		}

		int nNewLen = pDelKey - pTop;
		pHead->Add( pTop, nNewLen );
		delete m_pHeap;
		m_pHeap = pHead;
		
		SeekToFirst();
		return 1;
	}

	// in the middle
	else
	{
		MyVarMem* pHead = new MyVarMem;
		if( NULL == pHead )
		{
			return 0;
		}
		
		int nHeadLen = pDelKey - pTop;
		pHead->Add(pTop, nHeadLen);

		int nTailLen = nTotalLen - (pNext - pTop);
		pHead->Add(pNext, nTailLen);

		delete m_pHeap;
		m_pHeap = pHead;
		
		SeekToFirst();
		return 1;
	}
}

//-----------------------------------------------------------------------------
//	@detail
//		Get value
//	@param
//		char*	pszKey	key
//	@return
//		 value
//-----------------------------------------------------------------------------
const char* MyHashEx::GetValue(const char* pszKey)
{
	const char* p;

	p = SeekToKey(pszKey);
	if( NULL == p )
		return NULL;

	// Go head of 'value' string
	while(*p++ != '\0');
	
	return p;
}

//-----------------------------------------------------------------------------
//	@detail
//		Get value by the key without regard to case
//	@param
//		char*	pszKey	key without regard to case
//
//	@return
//		 l
//-----------------------------------------------------------------------------
const char* MyHashEx::GetValueI(const char *pszKey)
{
	const char* p;

	p = SeekToKeyI(pszKey);
	if( NULL == p )
		return NULL;

	while(*p++ != '\0');
	
	return p;

}

//-----------------------------------------------------------------------------
//	@detail
//		Get value
//	@param
//		char*	pszKey	key
//				key must be pointer that returned by SeekFirst or SeekNext
//	@return
//		 value
//-----------------------------------------------------------------------------
const char* MyHashEx::GetValueBySeekKey(SEEKKEY pszKey)
{
	// Go head of 'value' string
	while(*pszKey++ != '\0');
	
	return pszKey;
}

//-----------------------------------------------------------------------------
//	@detail
//		Set seek pointer to a head.
//	@param
//	@return
//-----------------------------------------------------------------------------
SEEKKEY	MyHashEx::SeekToFirst()
{
	return (char*)(m_pHeap->GetBuffer());
}

//-----------------------------------------------------------------------------
//	@detail
//		Seek pointer to the specified key.
//	@param
//		SEEKKEY	pszKey	key
//	@return
//		when there is, return the key.
//		otherwise, return null.
//-----------------------------------------------------------------------------
SEEKKEY	MyHashEx::SeekToKey(const char* pszKey)
{
	// set pointer to head of buffer
	const char* p = (char*) m_pHeap->GetBuffer();

	if( NULL == p )
	{
		return NULL;
	}

	while(StrCmp(p, pszKey) != 0)
	{
		p = SeekNext(p);

		if( NULL == p )
			return NULL;
	}
	return p;
}
//-----------------------------------------------------------------------------
//	@detail
//		Seek pointer to the specified key.
//	@param
//		SEEKKEY	pszKey
//				with out regard to case
//	@return
//		when there is, return the key.
//		otherwise, return null.
//-----------------------------------------------------------------------------
SEEKKEY	MyHashEx::SeekToKeyI(const char* pszKey)
{
	SeekToFirst();
	const char* p = (char*) m_pHeap->GetBuffer();

	if( NULL == p )
	{
		return NULL;
	}

	while(StrCmpI(p, pszKey) != 0)
	{
		p = SeekNext(p);
		if( NULL == p )
			return NULL;
	}
	return p;
}

//-----------------------------------------------------------------------------
//	@detail
//		Seek pointer to next key.
//	@param
//		char*	pszKey
//				current key
//				key must be pointer that returned by SeekFirst or SeekNext
//	@return
//		Return the next key.
//-----------------------------------------------------------------------------
SEEKKEY MyHashEx::SeekNext(SEEKKEY pszKey)
{
	if( NULL == pszKey )
	{
		return NULL;
	}
	const char* p = pszKey;
		
	// Seek pointer to the current value.
	while(*p++ != '\0');
	
	// The value is ""
	if( p[0] == '\0' && p[1] == '\0' )
	{

		return NULL;
	}

	// Go the next.
	while(*p++ != '\0');

	// Return null if the end.
	if( *p == '\0' )
	{
		return NULL;
	}

	return p;
}

//-----------------------------------------------------------------------------
//	@detail
//		Get internal buffer
//	@param
//		noen
//	@return
//		none
//-----------------------------------------------------------------------------
void* MyHashEx::GetBuffer() const
{
	return m_pHeap->GetBuffer();
}
